import vibe.d;
import std.stdio, std.typecons;
import db, gds;
import vibe.core.log;
import gen;

DataBase database;
Gds gds_wrapper;

Tuple !(Ticket, GdsTicketInfo) [string] cache;

string make_ticket_snapshot (GdsTicketInfo info) {
	return info.Carrier ~ info.Flight ~ info.Date ~ info.Class;
}

@property bool empty (string s) { return s == ""; }

void searchPage (HttpServerRequest req, HttpServerResponse res) {
	logInfo ("/search '%s'", req.query);

	string date = req.query.get ("d", "");
	if (date.empty) {
		logError ("Search flight with empty date");
		throw new HttpStatusException (400, "Invalid date");
	}

	string from = req.query.get ("f", "");
	string to = req.query.get ("t", "");

	logInfo ("Searching flights ...");
	auto flights = gds_wrapper.FindFlights (date, from, to);
	if (flights is null) {
		logError ("Incorrect search parameters");
		throw new HttpStatusException (400, "Invalid parameters");
	}

	res.setCookie ("viewid", "");

	res.renderCompat !(
		"flights.dt",
		string, "from",
		string, "to",
		string, "date",
		GdsFlightInfo [], "flights")
		(Variant (from), Variant (to), Variant (date), Variant (flights));
}

void reserveFlightPage (HttpServerRequest req, HttpServerResponse res) {
	logInfo ("/reserve '%s'", req.form);

	string name = req.form.get ("name", "");
	if (name.empty) {
		logError ("Reserve flight with empty name");
		throw new HttpStatusException (400, "Invalid name");
	}

	string surname = req.form.get ("surname", "");
	if (surname.empty) {
		logError ("Reserve flight with empty surname");
		throw new HttpStatusException (400, "Invalid surname");
	}

	string from = req.form.get ("from", "");
	if (from.empty) {
		logError ("Reserve flight with empty 'from' airport");
		throw new HttpStatusException (400, "Invalid 'from'");
	}

	string to = req.form.get ("to", "");
	if (to.empty) {
		logError ("Reserve flight with empty 'to' airport");
		throw new HttpStatusException (400, "Invalid 'to'");
	}

	string date = req.form.get ("date", "");
	if (date.empty) {
		logError ("Reserve flight with empty flight date");
		throw new HttpStatusException (400, "Invalid date");
	}

	string carrier = req.form.get ("carrier", "");
	if (carrier.empty) {
		logError ("Reserve flight with empty 'carrier'");
		throw new HttpStatusException (400, "Invalid carrier");
	}

	string flight = req.form.get ("flight", "");
	if (flight.empty) {
		logError ("Reserve flight with empty 'flight'");
		throw new HttpStatusException (400, "Invalid flight");
	}

	string clazz = req.form.get ("class", "");
	if (clazz.empty) {
		logError ("Reserve flight with empty 'class'");
		throw new HttpStatusException (400, "Invalid class");
	}

	auto reservationId = gds_wrapper.DoReservation (carrier, flight, clazz, date, from, to, name, surname);
	if (reservationId.empty) {
		logError ("Can't do reservation");
		throw new HttpStatusException (400, "Can't do reservation");
	}

	res.setCookie ("viewid", "");

	res.writeBody (xformat ("RESERVATION ID: %s", reservationId));
}

void viewFlightPage (HttpServerRequest req, HttpServerResponse res) {
	logInfo ("/viewflight '%s'", req.query);

	string date = req.query.get ("d", "");
	if (date.empty) {
		logError ("View flight with empty date");
		throw new HttpStatusException (400, "Invalid date");
	}

	string from = req.query.get ("f", "");
	string _to = req.query.get ("t", "");

	int flightNumber = to !int (req.query.get ("n", "-1"));
	if (flightNumber < 0) {
		logError ("View flight with empty flight#");
		throw new HttpStatusException (400, "Invalid flight#");
	}

	auto flights = gds_wrapper.FindFlights (date, from, _to);
	if (flights is null) {
		logError ("Incorrect search parameters");
		throw new HttpStatusException (400, "Invalid parameters");
	}

	if (flightNumber >= flights.length) {
		logError ("View flight with too large flight#");
		throw new HttpStatusException (400, "Flight# is out of range");
	}

	auto flight = flights [flightNumber];

	res.setCookie ("viewid", "");

	res.renderCompat !(
		"flight.dt",
		string, "date",
		GdsFlightInfo, "flight")
		(Variant (date), Variant (flight));
}

void buyTicketHandler (HttpServerRequest req, HttpServerResponse res) {
	logInfo ("/buy '%s'", req.query);

	string reservation = req.query.get ("r", "");
	if (reservation.empty) {
		logError ("Buy ticket with empty reservation#");
		throw new HttpStatusException (400, "Invalid reservation#");
	}

	string card = req.query.get ("c", "");
	if (card.empty) {
		logError ("Buy ticket with empty card#");
		throw new HttpStatusException (400, "Invalid card#");
	}

	string seat = req.query.get ("s", "");
	if (seat.empty) {
		logError ("Buy ticket with empty seat#");
		throw new HttpStatusException (400, "Invalid seat#");
	}

	logInfo ("Try to buy ticket with res#:'%s', card#:'%s', seat#:'%s'", reservation, card, seat);

	logInfo ("Checking double-purchase ...");
	if (database.Check (reservation)) {
		logError ("Ticket already purchased");
		throw new HttpStatusException (409, "Ticket already reserved");
	}
	logInfo ("OK");

	logInfo ("Checking reservation ...");
	if (! gds_wrapper.CheckReservation (reservation)) {
		logError ("Ticket not booked");
		throw new HttpStatusException (404, "Reservation not found");
	}
	logInfo ("OK");

	logInfo ("Retrieving information about reservation ...");
	auto info = make_ticket_snapshot (gds_wrapper [reservation]);
	if (! database.CheckSeatIsEmpty (info, seat)) {
		logError ("Seat is not empty");
		throw new HttpStatusException (409, "Seat is busy");
	}

	logInfo ("Storing ticket ...");
	database.Push (reservation, card, 10, info, seat);

	res.setCookie ("viewid", "");

	res.redirect (xformat ("/view?r=%s", reservation));
}

void changeTicketPage (HttpServerRequest req, HttpServerResponse res) {
	logInfo ("/change '%s'", req.query);

	string reservation = req.query.get ("r", "");
	if (reservation.empty) {
		logError ("Change ticket with empty reservation#");
		throw new HttpStatusException (400, "Invalid reservation#");
	}

	uint weight = to !uint (req.query.get ("w", "0"));
	if (0 == weight) {
		logError ("Change ticket to empty luggage weight");
		throw new HttpStatusException (400, "Invalid weight");
	}

	logInfo ("Try to change ticket with res#:'%s' (w:'%d')", reservation, weight);

	logInfo ("Checking reservation ...");
	if (! database.Check (reservation)) {
		logError ("Ticket is not purchased");
		throw new HttpStatusException (404, "Ticket not found");
	}
	logInfo ("OK");

	logInfo ("Changing ticket ...");
	auto index = database [reservation];

	if (index == -1) {
		logError ("Oops… All broken");
		throw new HttpStatusException (500, "All broken");
	}

	database.Change (index, weight);

	res.setCookie ("viewid", "");

	res.redirect (xformat ("/view?r=%s", reservation));
}

void viewTicketPage (HttpServerRequest req, HttpServerResponse res) {
	logInfo ("/view '%s'", req.query);

	string reservation = req.query.get ("r", "");
	if (reservation.empty) {
		logError ("View ticket with empty reservation#");
		throw new HttpStatusException (400, "Invalid reservation#");
	}

	Ticket ticket;
	GdsTicketInfo ticket_info;
	string sid;

	if ("viewid" !in req.cookies || req.cookies ["viewid"] == "") {
		logInfo ("Generating session id");
		sid = Generator.Next ();
		res.setCookie ("viewid", sid);
		goto label1;
	}

	if (req.cookies ["viewid"] in cache) {
		logInfo ("Using cache ...");
		sid = req.cookies ["viewid"];
		ticket = cache [sid] [0];
		ticket_info = cache [sid] [1];
		goto label2;
	}

label1:
	logInfo ("Checking reservation ...");
	if (! database.Check (reservation)) {
		logError ("Ticket is not purchased");
		throw new HttpStatusException (404, "Ticket not found");
	}
	logInfo ("OK");

	auto index = database [reservation];
	logInfo ("Ticket index is '%d'", index);

	ticket = database [index];
	if (ticket is null) {
		logError ("Oops");
		throw new HttpStatusException (500, "Oops… All broken :'(");
	}

	ticket_info = gds_wrapper [reservation];

	cache [sid] = tuple (ticket, ticket_info);

label2:
	logInfo ("View ticket with res#:'%s'", reservation);
	res.renderCompat !(
		"ticket.dt",
		Ticket, "ticket",
		GdsTicketInfo, "gds_ticket")
		(Variant (ticket), Variant (ticket_info));
}

void returnTicketHandler (HttpServerRequest req, HttpServerResponse res) {
	logInfo ("/return '%s'", req.query);

	string reservation = req.query.get ("r", "");
	if (reservation.empty) {
		logError ("Return ticket with empty reservation#");
		throw new HttpStatusException (400, "Invalid reservation#");
	}

	logInfo ("Checking reservation ...");
	if (! database.Check (reservation)) {
		logError ("Ticket is not purchased");
		throw new HttpStatusException (404, "Ticket not found");
	}
	logInfo ("OK");

	database.Remove (database [reservation]);
	gds_wrapper.RemoveReservation (reservation);

	res.setCookie ("viewid", "");

	res.redirect ("/");
}

void indexPage (HttpServerRequest req, HttpServerResponse res) {
	res.setCookie ("viewid", "");
	
	res.redirect ("index.htm");
}

void errorPage (HttpServerRequest req, HttpServerResponse res, HttpServerErrorInfo err) {
	res.renderCompat !(
		"error.dt",
		HttpServerRequest, "req",
		HttpServerErrorInfo, "err")
		(Variant (req), Variant (err));
}

static this () { 
	setLogFile ("booking.log", LogLevel.Info);

	database = new DataBase;
	gds_wrapper = new Gds ();

	auto router = new UrlRouter;
	router
	.get ("/", & indexPage)
	.get ("/search", & searchPage)
	.get ("/flight", & viewFlightPage)
	.post ("/reserve", & reserveFlightPage)
	.get ("/buy", & buyTicketHandler)
	.get ("/change", & changeTicketPage)
	.get ("/view", & viewTicketPage)
	.get ("/return", &returnTicketHandler)
	.get ("*", serveStaticFiles ("./public/"));

	auto settings = new HttpServerSettings;
	with (settings) {
		port = 8000;
		options = HttpServerOption.ParseQueryString | HttpServerOption.ParseCookies | HttpServerOption.ParseFormBody;
		errorPageHandler = toDelegate (& errorPage);
	}
	listenHttp (settings, router);
}

