module db;

import std.file, std.stdio, std.string;
import vibe.core.log;

const db_folder = "./db/";

class Ticket {
private:
	string reservationId;
	string cardId;
	string snapshot;
	uint luggage_weight;
	string seatNo;

public:
	this (string resId, File file) {
		this.reservationId = resId;
		file.readf ("%s\n%s\n%d\n%s", & cardId, & snapshot, & luggage_weight, & seatNo);
	}

	this (string resId, string card, uint weight, string snapshot, string seat) {
		this.reservationId = resId;
		this.cardId = card;
		this.luggage_weight = weight;
		this.snapshot = snapshot;
		this.seatNo = seat;
	}

	void Dump (File file) {
		file.writeln (this.cardId);
		file.writeln (this.snapshot);
		file.writeln (this.luggage_weight);
		file.writeln (this.seatNo);
	}

	@property string ReservationId () { return this.reservationId; }

	@property string CardId () { return this.cardId; }

	@property uint LuggageWeight () { return this.luggage_weight; }
	@property uint LuggageWeight (uint weight) { return this.luggage_weight = weight; }

	@property string SeatNumber () { return this.seatNo; }
}

static if (is (typeof ([0].length) == uint)) {
	alias int Index;
}
else {
	alias long Index;
}

class DataBase {
private:
	Ticket [] tickets;
	Index [string] rid_index;

	static void InitializeEmptyDatabase () {
		if (exists (db_folder))
			return;

		logWarn ("Database folder not exists. Creating ...");
		mkdir (db_folder);
		logInfo ("Database folder created");
	}

	void Load (string id) {
		if (Check (id))
			return ;

		auto name = db_folder ~ id;
		logInfo ("Import data from '%s' file ...", name);

		if (! exists (name)) {
			logError ("File not found");
			return;
		}

		auto f = File (name, "r");
		this.tickets ~= new Ticket (id, f);
		this.rid_index [id] = this.tickets.length - 1;
		f.close ();

		logInfo ("Loading complete");
	}

	void Store (Index index) {
		auto ticket = this.tickets [index];

		auto file = File (db_folder ~ ticket.reservationId, "w");
		ticket.Dump (file);
		file.close ();
	}

public:
	this () {
		logInfo ("Initializing database ...");

		logInfo ("Initializing database folder ...");
		InitializeEmptyDatabase ();
		logInfo ("Done");

		logInfo ("Import data ...");

		auto entries = dirEntries (db_folder, SpanMode.shallow);

		foreach (string name ; entries) {
			logInfo ("Reading file '%s'", name);
			auto file = File (name, "r");
			string s = name [name.lastIndexOf ("/") + 1 .. $];

			this.tickets ~= new Ticket (s, file);
			this.rid_index [s] = this.tickets.length - 1;

			file.close ();
		}
		logInfo ("Done");
	}

	bool Check (string id) {
		logInfo ("Checking ticket with id# = '%s'", id);

		scope (exit) logInfo ("Complete");

		foreach (Index i, Ticket t ; this.tickets)
			if (t !is null && t.reservationId == id)
				return true;

		return false;
	}

	bool CheckSeatIsEmpty (string snapshot, string seat) {
		logInfo ("Checking seat '%s' for '%s'", seat, snapshot);

		scope (exit) logInfo ("Complete");

		foreach (Index i, Ticket t ; this.tickets)
			if (t !is null && t.snapshot == snapshot && t.seatNo == seat)
				return false;

		return true;
	}

	void Push (string id, string cardId, uint weight, string snapshot, string seat) {
		logInfo ("Add '%s' to '%s' (%d, %s%s)", id, cardId, weight, snapshot, seat);

		this.tickets ~= new Ticket (id, cardId, weight, snapshot, seat);
		this.rid_index [id] = this.tickets.length - 1;

		Store (this.tickets.length - 1);
		logInfo ("Complete");
	}

	void Remove (Index index) {
		logInfo ("Removing #%d", index);

		if (index < 0 || index >= this.tickets.length) {
			logError ("Incorrect Id");
			return;
		}

		scope (exit) logInfo ("Complete");

		if (this.tickets.length == 1) {
			this.tickets = [ ];
			return;
		}

		if (index == 0) {
			this.tickets = this.tickets [1 .. $];
			return;
		}

		if (index == this.tickets.length - 1) {
			this.tickets = this.tickets [0 .. $-1];
			return;
		}

		auto id = this.tickets [index].reservationId;
		remove (db_folder ~ id);
		this.tickets = this.tickets [0 .. index] ~ this.tickets [index + 1 .. $];
		this.rid_index.remove (id);

		foreach (Index i, Ticket t ; this.tickets)
			if (t !is null)
				this.rid_index [t.reservationId] = i;
	}

	void Change (Index index, uint weight) {
		logInfo ("Changing #%d (%d)", index, weight);

		if (index < 0 || index >= this.tickets.length) {
			logError ("Incorrect Id");
			return;
		}

		this.tickets [index].LuggageWeight = weight;
		Store (index);

		logInfo ("Complete");
	}

	void Clean () {
		logInfo ("Cleaning database ...");

		InitializeEmptyDatabase ();

		logInfo ("Removing data ...");
		foreach (string name ; dirEntries (db_folder, SpanMode.shallow)) {
			if (! name.isFile) {
				logWarn ("'%s' is not a database entry", name);
				continue;
			}

			scope (success) logInfo ("Remove OK");
			scope (failure) logError ("Remove FAILED");

			logInfo ("Remove entry '%s'", name);
			remove (name);
		}

		logInfo ("Cleaning complere");
	}

	Ticket opIndex (Index index) {
		if (index < 0 || index >= this.tickets.length) {
			logError ("Incorrect Id");
			return null;
		}

		return this.tickets [index];
	}

	Index opIndex (string id) {
		Load (id);

		if (id in this.rid_index)
			return this.rid_index [id];

		return -1;
	}
}

