module gds;

import std.stdio, std.conv, std.exception;
import std.regex, std.array, std.algorithm;
import std.stream, std.socket, std.socketstream;
import vibe.core.log;

class GdsTicketInfo {
private:
	string name;
	string surname;
	string carrier;
	string flight;
	string clazz;
	string date;
	string from;
	string to;

public:
	this () { }

	this (string info) {
		auto strings = info.split (regex ("\\s+"));

		this.name = strings [1];
		this.surname = strings [2];
		this.carrier = strings [6];
		this.flight = strings [7];
		this.clazz = strings [8];
		this.date = strings [9];
		this.from = strings [11] [0 .. 3];
		this.to = strings [11] [3 .. 6];
	}

	@property string Name () { return this.name; }

	@property string Surname () { return this.surname; }

	@property string Carrier () { return this.carrier; }

	@property string Flight () { return this.flight; }

	@property string Class () { return this.clazz; }

	@property string Date () { return this.date; }

	@property string From () { return this.from; }

	@property string To () { return this.to; }
}

class GdsFlightInfo {
private:
	string carrier;
	string flight;
	string [string] seats;
	string from;
	string to;
	string time_departure;

public:
	this (string info) {
		auto strings = info.split (regex ("\\s+"));

		this.carrier = strings [1];
		this.flight = strings [2];
		foreach (int i, string v ; strings [3 .. $ - 3])
			this.seats [v [0 .. 1]] = v [1 .. $];
		this.from = strings [$ - 3];
		this.to = strings [$ - 2];
		this.time_departure = strings [$ - 1];
	}

	@property string Carrier () { return this.carrier; }

	@property string Flight () { return this.flight; }

	@property string [] Classes () { return this.seats.keys; }

	string SeatsCount (string clazz) {
		return this.seats.get (clazz, "0");
	}

	@property string From () { return this.from; }

	@property string To () { return this.to; }

	@property string DepartureTime () { return this.time_departure; }
}

private @property IsError (char [] s) {
	return s.startsWith ("CHECK YOUR INPUT");
}

class Gds {
private:
	Socket socket;
	Stream socket_stream;

	string login;
	string password;
	ushort port;

	void load_config () {
		logInfo ("Loading config ...");
		auto f = std.stdio.File ("config", "r");
		f.readf ("%s\n%s\n%d", & this.login, & this.password, & this.port);
		f.close ();
		logInfo ("OK");
	}

	bool Reconnect () {
		if (this.socket !is null)
			return true;

		logInfo ("Trying reconnect ...");

		auto result = false;

		try {
			this.socket = new TcpSocket (new InternetAddress ("127.0.0.1", this.port));
			this.socket_stream = new SocketStream (this.socket);
			this.socket_stream.writeLine (join (["JI", this.login, this.password], " "));

			auto answer = this.socket_stream.readLine ();
			result = answer.startsWith ("** WELCOME");
		}
		catch (Exception e) {
			return false;
		}

		if (! result)
			this.socket = null;

		return result;
	}

public:
	this () {
		this.load_config ();

		logInfo ("Connecting to GDS ...");
		this.Reconnect ();
		logInfo ("Connected");
	}

	GdsFlightInfo [] FindFlights (string date, string from, string _to) {
		if (! this.Reconnect ())
			return null;

		this.socket_stream.writeLine (join (["A", "N", date, from, _to], " "));

		auto answer = this.socket_stream.readLine ();
		if (answer.IsError)
			return null;

		GdsFlightInfo [] _result = [ ];
		auto result = appender (_result);

		auto ans = to !string (this.socket_stream.readLine ());
		try {
			while (ans != "") {
				result.put (new GdsFlightInfo (ans));

				ans = to !string (this.socket_stream.readLine ());
			}
		}
		catch (Exception e) { }

		return result.data;
	}

	string DoReservation (
		string carrier,
		string flight,
		string clazz,
		string date,
		string from,
		string _to,
		string name,
		string surname) {
		if (! this.Reconnect ())
			return "";

		this.socket_stream.writeLine (join (["SS", carrier, flight, clazz, date, from, _to, "1"], " "));
		auto answer = this.socket_stream.readLine ();
		if (answer.IsError)
			return "";

		this.socket_stream.writeLine (join (["NM", surname, name, "Mr"], " "));
		this.socket_stream.readLine ();
		if (answer.IsError)
			return "";

		this.socket_stream.writeLine ("ET");

		answer = this.socket_stream.readLine ();
		if (answer.startsWith ("END OF TRANSACTION COMPLETE"))
			return to !string (answer.split (regex (" ")) [$ - 1]);

		return "";
	}

	bool CheckReservation (string reservation) {
		if (! this.Reconnect ())
			return false;

		this.socket_stream.writeLine ("RT " ~ reservation);

		auto answer = this.socket_stream.readLine ();
		if (answer.IsError || answer.startsWith ("NOT FOUND"))
			return false;

		return true;
	}

	GdsTicketInfo opIndex (string reservation) {
		if (! this.Reconnect ())
			return null;

		this.socket_stream.writeLine ("RT " ~ reservation);

		auto answer = this.socket_stream.readLine ();
		if (answer.IsError || answer.startsWith ("NOT FOUND"))
			return null;

		return new GdsTicketInfo (to !string (answer));
	}

	void RemoveReservation (string reservation) {
		if (! this.Reconnect ())
			return ;

		this.socket_stream.writeLine ("DE " ~ reservation);
	}

	~this () {
		this.socket_stream.writeLine ("JO");
		this.socket.close ();
	}
}

